// Copyright © 2020 Amazon Web Services
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.aws.config;

import com.amazonaws.auth.AWSCredentialsProvider;
import com.amazonaws.services.kms.AWSKMS;
import com.amazonaws.services.simplesystemsmanagement.AWSSimpleSystemsManagement;
import com.amazonaws.services.simplesystemsmanagement.AWSSimpleSystemsManagementClientBuilder;
import com.amazonaws.services.simplesystemsmanagement.model.GetParametersRequest;
import com.amazonaws.services.simplesystemsmanagement.model.GetParametersResult;
import com.amazonaws.services.simplesystemsmanagement.model.ParameterNotFoundException;
import com.amazonaws.services.simplesystemsmanagement.model.InternalServerErrorException;
import com.amazonaws.services.simplesystemsmanagement.model.Parameter;
import com.amazonaws.services.sns.AmazonSNS;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.Setter;
import org.apache.http.HttpStatus;
import org.opengroup.osdu.core.aws.iam.IAMConfig;
import org.opengroup.osdu.core.aws.sns.AmazonSNSConfig;
import org.opengroup.osdu.register.provider.aws.subscriber.KmsConfig;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import javax.annotation.PostConstruct;
import java.util.ArrayList;
import java.util.List;
import org.opengroup.osdu.core.common.model.http.AppException;

@Component
public class AwsServiceConfig {

  @Value("${aws.region}")
  @Getter()
  @Setter(AccessLevel.PROTECTED)
  public String amazonRegion;

  @Value("${aws.ssm}")
  @Getter()
  @Setter(AccessLevel.PROTECTED)
  public Boolean ssmEnabled;

  @Value("${aws.kms.endpoint}")
  @Getter()
  @Setter(AccessLevel.PROTECTED)
  public String kmsEndpoint;


  @Value("${aws.environment}")
  @Getter()
  @Setter(AccessLevel.PROTECTED)
  public String environment;

  @Getter()
  public String kmsKeyId;

  @Getter()
  public String snsTopicArn;

  @Getter()
  public AWSKMS kmsClient;

  @Getter()
  public AmazonSNS snsClient;

  private AWSCredentialsProvider amazonAWSCredentials;
  private AWSSimpleSystemsManagement ssmManager;

  @PostConstruct
  public void init() {
    if (ssmEnabled) {

      AmazonSNSConfig snsConfig = new AmazonSNSConfig(amazonRegion);
      snsClient = snsConfig.AmazonSNS();
      KmsConfig kmsConfig = new KmsConfig(kmsEndpoint, amazonRegion);
      kmsClient = kmsConfig.awsKMS();

      try {
        String keyssmparameter = "/osdu/" + environment + "/register/kpath";
        String snstopicArnParameter = "/osdu/" + environment + "/register/register-sns-topic-arn";
        amazonAWSCredentials = IAMConfig.amazonAWSCredentials();
        ssmManager = AWSSimpleSystemsManagementClientBuilder.standard()
                .withCredentials(amazonAWSCredentials)
                .withRegion(amazonRegion)
                .build();
        List<String> paramsList = new ArrayList<String>();
        paramsList.add(keyssmparameter);
        paramsList.add(snstopicArnParameter);

        GetParametersRequest paramRequest = new GetParametersRequest()
                .withNames(paramsList)
                .withWithDecryption(true);
        GetParametersResult paramResult = new GetParametersResult();
        paramResult = ssmManager.getParameters(paramRequest);
        System.out.println("SSM Parameters retrieved");
        List<Parameter> paramsResultList = new ArrayList<>();
        paramsResultList = paramResult.getParameters();
        for (Parameter s : paramsResultList) {
          if(s.getName().equalsIgnoreCase(keyssmparameter)) {
            kmsKeyId = s.getValue();
          }
          if(s.getName().equalsIgnoreCase(snstopicArnParameter)) {
            snsTopicArn = s.getValue();
          }

        }
      } catch (ParameterNotFoundException e) {
        throw new AppException(HttpStatus.SC_NOT_FOUND, "SSM ParameterNotFoundException",e.getErrorMessage());
      } catch (InternalServerErrorException e) {
        throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "SSM InternalServerErrorException",e.getErrorMessage());

      }
    }
  }


}

