// Copyright © 2020 Amazon Web Services
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.aws.action;


import com.amazonaws.services.dynamodbv2.datamodeling.DynamoDBDeleteExpression;
import com.amazonaws.services.dynamodbv2.model.ConditionalCheckFailedException;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.runners.MockitoJUnitRunner;


import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperFactory;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperV2;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.action.model.Action;
import org.opengroup.osdu.register.provider.aws.util.TestUtils;

import java.util.ArrayList;

import java.util.List;


import static org.junit.Assert.assertEquals;


import static org.mockito.Mockito.lenient;
import static org.mockito.MockitoAnnotations.initMocks;

@RunWith(MockitoJUnitRunner.class)
public class AwsActionRepoTest {

    @Mock
    private DpsHeaders dpsHeaders;

    @Mock
    private DynamoDBQueryHelperV2 queryHelper;

    @Mock
    private DynamoDBQueryHelperFactory queryHelperFactory;

    @Mock
    private JaxRsDpsLog logger;

    @InjectMocks
    private AwsActionRepo repo;


    @Before
    public void setUp() {
        lenient().doReturn(TestUtils.getDataPartitionId()).when(dpsHeaders).getPartitionId();

        initMocks(this);
        Mockito.when(queryHelperFactory.getQueryHelperForPartition(Mockito.any(DpsHeaders.class), Mockito.any()))
                .thenReturn(queryHelper);
    }

    @Test
    public void createAction_success(){
        Action action = createMockAction();
        Mockito.doNothing().when(queryHelper).saveWithHashCondition(Mockito.any(ActionDoc.class),Mockito.anyString());
        Action resultAction = repo.createAction(action);
        assertEquals(action, resultAction);

    }


    @Test
    public void createAction_throw400_whenIdNull(){
        Action action = createMockAction();
        action.setId(null);

        try {
             repo.createAction(action);
        } catch (AppException e) {
            assertEquals(400, e.getError().getCode());
        }
    }

    @Test
    public void createAction_throw409_whenIdAlreadyExists(){
        Action action = createMockAction();
        AppException e = new AppException(409, "Conflict", String.format("An action already exists with the id: %s", action.getId()));
        Mockito.doThrow(e).when(queryHelper).saveWithHashCondition(Mockito.any(ActionDoc.class),Mockito.anyString());

        try {
         repo.createAction(action);
        } catch (AppException ex) {
            assertEquals(409, ex.getError().getCode());
        }
    }

    @Test
    public void deleteAction_success(){


        String id = TestUtils.getAction_id();
        Mockito.doNothing().when(queryHelper).deleteByObjectWithCondition(Mockito.any(ActionDoc.class),Mockito.any(DynamoDBDeleteExpression.class));
        Boolean result = repo.delete(id);
        assertEquals(result, true);
    }

    @Test
    public void deleteAction_throw404_whenIdoesNotExist(){
        String id = TestUtils.getAction_id();
        ConditionalCheckFailedException e = new ConditionalCheckFailedException("Failed to delete");
        Mockito.doThrow(e).when(queryHelper).deleteByObjectWithCondition(Mockito.any(ActionDoc.class),Mockito.any(DynamoDBDeleteExpression.class));

        try {
              repo.delete(id);
        } catch (AppException ex) {
            assertEquals(404, ex.getError().getCode());

        }
    }

    @Test
    public void getAction_success(){
        Action action = createMockAction();
        String id = TestUtils.getAction_id();
        ActionDoc doc = ActionDoc.mapFrom(action, TestUtils.getDataPartitionId());
        Mockito.when(queryHelper.loadByPrimaryKey(ActionDoc.class, id)).thenReturn(doc);

        Action resultAction = repo.get(id);
        action.setCreatedOnEpoch(resultAction.getCreatedOnEpoch());
        assertEquals(action, resultAction);

    }

    @Test
    public void getAction_throw404_whenIdoesNotExist(){
        String id = TestUtils.getAction_id();
        AppException e = new AppException(404,"","");
        Mockito.doThrow(e).when(queryHelper).loadByPrimaryKey(ActionDoc.class, id);
        try {
            repo.get(id);
        } catch (AppException ex) {
            assertEquals(404, ex.getError().getCode());

        }
    }

    @Test
    public void getAllActions_success(){
        Action action1 =  createMockAction();
        Action action2 =  createMockAction();
        action2.setId("ActionId2");

        ActionDoc doc1 = ActionDoc.mapFrom(action1,TestUtils.getDataPartitionId());
        ActionDoc doc2 = ActionDoc.mapFrom(action2,TestUtils.getDataPartitionId());

        ArrayList<ActionDoc> actionDocs = new ArrayList<>();
        actionDocs.add(doc1);
        actionDocs.add(doc2);
        Mockito.when(queryHelper.scanTable(Mockito.any(Class.class),Mockito.anyString(),Mockito.anyMap())).thenReturn(actionDocs);
        List<Action> resultActions = repo.getAllActions();
        action1.setCreatedOnEpoch(resultActions.get(0).getCreatedOnEpoch());
        action2.setCreatedOnEpoch(resultActions.get(1).getCreatedOnEpoch());
        List<Action> actions =new ArrayList<>();
        actions.add(action1);
        actions.add(action2);
        assertEquals(actions, resultActions);
    }

    private Action createMockAction() {
        Action action = new Action();
        action.setId(TestUtils.getAction_id());
        action.setName(TestUtils.getAction_name());
        action.setDescription(TestUtils.getAction_description());
        action.setImg(TestUtils.getAction_img());
        action.setUrl(TestUtils.getAction_url());
        action.setContactEmail(TestUtils.getAction_contactEmail());
        action.setFilter(TestUtils.getAction_filter());

        return action;
    }


}
