// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.subscriber;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.jupiter.MockitoExtension;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.register.provider.azure.subscriber.pubsub.IPubsubSubscription;
import org.opengroup.osdu.register.subscriber.model.HmacSecret;
import org.opengroup.osdu.register.subscriber.model.Secret;
import org.opengroup.osdu.register.subscriber.model.Subscription;

import java.util.Collections;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

@ExtendWith(MockitoExtension.class)
public class SubscriptionRepositoryTest {

    private static final String subscriptionId = "id-1";
    private static final String notificationId = "id-2";
    private static final String secretValue = "some-value";
    private static final String topicName = "topic-name";

    @Mock
    private IPubsubSubscription pubSubSubscription;

    @Mock
    private CosmosSubscription cosmosSubscription;

    @Mock
    private JaxRsDpsLog logger;

    @Mock
    private Subscription subscription;

    @InjectMocks
    private SubscriptionRepository repo;

    @BeforeEach
    void init() {
        lenient().when(subscription.getId()).thenReturn(subscriptionId);
        lenient().when(subscription.getSecret()).thenReturn(new HmacSecret(secretValue));
        lenient().when(subscription.getNotificationId()).thenReturn(notificationId);
        lenient().when(subscription.getTopic()).thenReturn(topicName);
    }

    @Test
    public void createSubscriptionThrows409WhenIdExists() {
        doThrow(new AppException(409, "Reason", "Message")).when(cosmosSubscription).createSubscription(subscription);
        doReturn(subscription).when(cosmosSubscription).getSubscription(subscriptionId);
        doNothing().when(pubSubSubscription).checkIfSubscriptionExists(notificationId, topicName);

        AppException exception = assertThrows(AppException.class, () -> {
            repo.create(subscription);
        });

        assertEquals(409, exception.getError().getCode());
        assertEquals("Conflict", exception.getError().getReason());
        assertEquals("A subscriber already exists with the same topic and endpoint combination", exception.getError().getMessage());
        verify(subscription, times(2)).getId();
        verify(logger, times(1)).error("A subscriber already exists with the same topic and endpoint combination");
    }

    @Test
    public void createSubscriptionCosmosInsertedSuccessfullyAndPubSubSubscriptionCreated() {
        when(cosmosSubscription.createSubscription(subscription)).thenReturn(subscription);
        doNothing().when(pubSubSubscription).createSubscription(notificationId, topicName);

        Subscription output = repo.create(subscription);

        assertEquals(subscription, output);
        verify(pubSubSubscription, times(1)).createSubscription(notificationId, topicName);
        verify(logger, times(1)).debug("PubSub Subscription created with ID:" + subscription.getNotificationId());
    }

    @Test
    public void createCosmosSubscriptionInsertedSuccessfullyButPubSubSubscriptionCreationFailed() {
        when(cosmosSubscription.createSubscription(subscription)).thenReturn(subscription);
        doNothing().when(cosmosSubscription).deleteSubscription(subscriptionId);
        doThrow(new AppException(500, "Server Error", "Unexpected error creating Push subscription")).when(pubSubSubscription).createSubscription(notificationId, topicName);

        AppException exception = assertThrows(AppException.class, () -> {
            repo.create(subscription);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        assertEquals("Server Error", exception.getError().getReason());
        assertEquals("Unexpected error creating subscription", exception.getError().getMessage());

        verify(subscription, times(1)).getId();
        verify(pubSubSubscription, times(1)).createSubscription(notificationId, topicName);
        verify(logger, times(1)).error("Unexpected error creating subscription");
    }

    @Test
    public void subscriptionCreatedSuccessfullyIfRecordExistsInCosmosButPubSubSubscriptionIsNotPresent() {
        doThrow(new AppException(409, "Reason", "Message")).when(cosmosSubscription).createSubscription(subscription);
        doThrow(new AppException(404, "Reason", "Message")).when(pubSubSubscription).checkIfSubscriptionExists(notificationId, topicName);
        when(cosmosSubscription.getSubscription(subscriptionId)).thenReturn(subscription);
        doNothing().when(cosmosSubscription).updateSubscription(subscription);

        Subscription output = repo.create(subscription);

        assertEquals(output, subscription);
        verify(subscription, times(2)).getId();
        verify(logger, times(1)).info("PubSub Subscription created with ID's:" + notificationId);
    }

    @Test
    public void createSubscriptionShouldThrow500ExceptionIfRecordExistsInCosmosButFailedToCreatePubSubSubscription() {
        doThrow(new AppException(409, "Reason", "Message")).when(cosmosSubscription).createSubscription(subscription);
        doThrow(new AppException(404, "Reason", "Message")).when(pubSubSubscription).checkIfSubscriptionExists(notificationId, topicName);
        doReturn(subscription).when(cosmosSubscription).getSubscription(subscriptionId);
        doThrow(new AppException(500, "Server Error", "Unexpected error creating Push subscription")).when(pubSubSubscription).createSubscription(notificationId, topicName);
        doNothing().when(cosmosSubscription).updateSubscription(subscription);

        AppException exception = assertThrows(AppException.class, () -> {
            repo.create(subscription);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        assertEquals("Server Error", exception.getError().getReason());
        assertEquals("Unexpected error creating subscription", exception.getError().getMessage());
        verify(subscription, times(2)).getId();
        verify(pubSubSubscription, times(1)).createSubscription(notificationId, topicName);
        verify(pubSubSubscription, times(1)).checkIfSubscriptionExists(notificationId, topicName);
        verify(logger, times(1)).error("Unexpected error creating subscription");
    }

    @Test
    public void createSubscriptionShouldThrow500ExceptionIfRecordExistsInCosmosButUpsertOperationFailed() {
        doThrow(new AppException(409, "Reason", "Message")).when(cosmosSubscription).createSubscription(subscription);
        doThrow(new AppException(404, "Reason", "Message")).when(pubSubSubscription).checkIfSubscriptionExists(notificationId, topicName);
        doThrow(new AppException(500, "Reason", "Message")).when(cosmosSubscription).updateSubscription(subscription);
        doReturn(subscription).when(cosmosSubscription).getSubscription(subscriptionId);
        AppException exception = assertThrows(AppException.class, () -> {
            repo.create(subscription);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        assertEquals("Server Error", exception.getError().getReason());
        assertEquals("Unexpected error creating subscription", exception.getError().getMessage());

        verify(subscription, times(2)).getId();
        verify(pubSubSubscription, times(1)).checkIfSubscriptionExists(notificationId, topicName);
        verify(logger, times(1)).error("Unexpected error creating subscription");
    }

    @Test
    public void createSubscriptionShouldThrow500ExceptionIfRecordExistsInCosmosButFailedToGetPubSubSubscription() {
        doThrow(new AppException(409, "Reason", "Message")).when(cosmosSubscription).createSubscription(subscription);
        doReturn(subscription).when(cosmosSubscription).getSubscription(subscriptionId);
        doThrow(new AppException(500, "Reason", "Message")).when(pubSubSubscription).checkIfSubscriptionExists(notificationId, topicName);

        AppException exception = assertThrows(AppException.class, () -> {
            repo.create(subscription);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        assertEquals("Server Error", exception.getError().getReason());
        assertEquals("Unexpected error creating subscription", exception.getError().getMessage());

        verify(subscription, times(2)).getId();
        verify(pubSubSubscription, times(1)).checkIfSubscriptionExists(notificationId, topicName);
        verify(logger, times(1)).error("Unexpected error creating subscription");
    }

    @Test
    public void getSubscriptionThrows404WhenIdNotFound() {
        when(cosmosSubscription.getSubscription(subscriptionId)).thenThrow(new AppException(404, "Not found", "Subscriber with id id-1 does not exist"));

        AppException exception = assertThrows(AppException.class, () -> {
            repo.get(subscriptionId);
        });

        assertEquals(404, exception.getError().getCode());
        assertEquals("Not found", exception.getError().getReason());
        assertEquals("Subscriber with id id-1 does not exist", exception.getError().getMessage());
    }

    @Test
    public void getSubscriptionItemRetrievedSuccessfully() {
        doReturn(subscription).when(cosmosSubscription).getSubscription(subscriptionId);

        Subscription output = repo.get(subscriptionId);

        assertEquals(subscription, output);
    }

    @Test
    public void querySubscriptionsReturnItemsSuccessfully() {
        Subscription expected = getSubscription();
        List<Subscription> subscriptionList = Collections.singletonList(expected);
        doReturn(subscriptionList).when(cosmosSubscription).querySubscription(notificationId);

        List<Subscription> output = repo.query(notificationId);
        expected.setCreatedOnEpoch(output.get(0).getCreatedOnEpoch());

        assertEquals(Collections.singletonList(expected), output);
    }

    @Test
    public void deleteSubscriptionReturnsFalseWhenIdNotFound() {
        when(cosmosSubscription.getSubscription(subscriptionId)).thenThrow(new AppException(404, "Not found", String.format("Subscriber with id %s does not exist.", subscriptionId)));

        boolean output = repo.delete(subscriptionId);

        assertFalse(output);
        verify(logger, times(1)).info("Record not found for subscription with ID: " + subscriptionId);
    }

    @Test
    public void deleteSubscriptionReturnsTrueWhenDeletedSuccessfully() {
        when(cosmosSubscription.getSubscription(subscriptionId)).thenReturn(subscription);
        doNothing().when(cosmosSubscription).deleteSubscription(subscriptionId);
        doNothing().when(pubSubSubscription).deleteSubscription(notificationId, topicName);

        boolean output = repo.delete(subscriptionId);

        assertTrue(output);
        verify(logger, times(1)).debug("Record found for subscription with ID: " + subscriptionId);
        verify(logger, times(1)).debug("PubSub Subscription deleted with ID:" + notificationId);
    }

    @Test
    public void deleteSubscriptionThrows500ExceptionWhenDeletionFailedInCosmosDb() {
        when(cosmosSubscription.getSubscription(subscriptionId)).thenReturn(subscription);
        doNothing().when(pubSubSubscription).deleteSubscription(notificationId, topicName);
        doThrow(new AppException(500, "Server Error", "Unexpected error deleting subscription")).when(cosmosSubscription).deleteSubscription(subscriptionId);

        AppException exception = assertThrows(AppException.class, () -> {
            repo.delete(subscriptionId);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        assertEquals("Server Error", exception.getError().getReason());
        assertEquals("Unexpected error deleting subscription", exception.getError().getMessage());
        verify(logger, times(1)).error("Unexpected error deleting subscription");
    }

    @Test
    public void deleteSubscriptionThrows500ExceptionWhenDeletingPubSubSubscriptionFailed() {
        when(cosmosSubscription.getSubscription(subscriptionId)).thenReturn(subscription);
        doThrow(new AppException(500, "Server Error", "Unexpected error deleting subscription")).when(pubSubSubscription).deleteSubscription(notificationId, topicName);

        AppException exception = assertThrows(AppException.class, () -> {
            repo.delete(subscriptionId);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        assertEquals("Server Error", exception.getError().getReason());
        assertEquals("Unexpected error deleting subscription", exception.getError().getMessage());
        verify(logger, times(1)).error("Unexpected error deleting subscription");
    }

    @Test
    public void deleteSubscriptionThrows409ExceptionWhenDeletePubSubSubscriptionHaveConflicts() {
        when(cosmosSubscription.getSubscription(subscriptionId)).thenReturn(subscription);
        doThrow(new AppException(409, "Conflict", "Another request is trying to delete the same subscription")).when(pubSubSubscription).deleteSubscription(notificationId, topicName);

        AppException exception = assertThrows(AppException.class, () -> {
            repo.delete(subscriptionId);
        });

        assertNotNull(exception);
        assertEquals(409, exception.getError().getCode());
        assertEquals("Conflict", exception.getError().getReason());
        assertEquals("Another request is trying to delete the same subscription", exception.getError().getMessage());
        verify(logger, times(1)).error("Another request is trying to delete the same subscription");
    }

    @Test
    public void deleteSubscriptionSuccessfullyWhenPubSubSubscriptionNotFoundAtDeletion() {
        when(cosmosSubscription.getSubscription(subscriptionId)).thenReturn(subscription);
        doThrow(new AppException(404, "Not Found", "Not Found message")).when(pubSubSubscription).deleteSubscription(notificationId, topicName);

        boolean output = repo.delete(subscriptionId);

        assertTrue(output);
        verify(logger, times(1)).debug("Record found for subscription with ID: " + subscriptionId);
        verify(logger, times(1)).info("PubSub Subscription does not exist for deletion");
    }

    @Test
    public void patchShouldSuccessfullyUpdateSecret() {
        Secret secret = Mockito.mock(Secret.class);
        when(cosmosSubscription.updateSecret(subscription, secret)).thenReturn(true);

        boolean output = repo.patch(subscription, secret);

        assertTrue(output);
    }

    @Test
    public void getAllSubscriptionsReturnItemsSuccessfully() {
        Subscription expected = getSubscription();
        List<Subscription> subscriptionList = Collections.singletonList(expected);
        doReturn(subscriptionList).when(cosmosSubscription).getAllSubscriptions();

        List<Subscription> output = repo.getAll();
        expected.setCreatedOnEpoch(output.get(0).getCreatedOnEpoch());

        assertEquals(Collections.singletonList(expected), output);

    }

    private Subscription getSubscription() {
        Subscription expected = new Subscription();
        expected.setSecret(new HmacSecret(secretValue));
        expected.setNotificationId(subscriptionId);
        expected.setTopic(topicName);
        return expected;
    }
}
