/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.provider.gcp.action.datastore;

import java.util.ArrayList;
import java.util.List;

import org.apache.http.HttpStatus;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.common.util.Crc32c;
import org.opengroup.osdu.register.action.model.Action;
import org.opengroup.osdu.register.provider.gcp.ddms.datastore.DatastoreMultiTenantAccess;
import org.opengroup.osdu.register.provider.interfaces.action.IActionRepo;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import com.google.cloud.Timestamp;
import com.google.cloud.datastore.Datastore;
import com.google.cloud.datastore.DatastoreException;
import com.google.cloud.datastore.Entity;
import com.google.cloud.datastore.FullEntity;
import com.google.cloud.datastore.Key;
import com.google.cloud.datastore.Query;
import com.google.cloud.datastore.QueryResults;
import com.google.cloud.datastore.StringValue;
import com.google.cloud.datastore.Transaction;
import com.google.cloud.datastore.Value;
import com.google.gson.Gson;

import lombok.Getter;

@Repository
public class DatastoreActionRepo implements IActionRepo {

    private static final String CONFLICT_REASON = "Conflict";

    @Getter
    private static final String NAMESPACE = "DE";
    @Getter
    private static final String KIND = "ACTION";
    @Getter
    private static final String NAME = "name";
    @Getter
    private static final String CREATED_ON = "createdOn";
    @Getter
    private static final String CONTACT_EMAIL = "contactEmail";
    private static final String DESCRIPTION = "description";
    private static final String URL = "url";
    private static final String IMG = "img";
    private static final String TYPE = "type";
    private static final String SOURCE = "source";
    private static final String VERSION = "version";
    @Getter
    private static final String CHECKSUM = "checkSum";

    @Autowired
    private JaxRsDpsLog log;
    @Autowired
    private DatastoreMultiTenantAccess dataStoreTenants;
    @Autowired
    private TenantInfo tenantInfo;

    @Override
    public Action createAction(Action a) {
        return createActionInDatastore(a);
    }

    @Override
    public Action get(String id) {
        Datastore ds = dataStoreTenants.get(tenantInfo);
        Key key = createkey(ds, id);
        Entity entity = ds.get(key);
        if (entity == null)
            throw new AppException(404, "Not found", String.format("Action with id %s does not exist.", id));
        Action action = convertEntityToAction(entity);
        if (getActionChecksum(action).compareTo(entity.getString(CHECKSUM)) != 0) {
            throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "Internal server error", String.format("Error retrieving action '%s'.", action.getName()));
        }
        return action;
    }

    @Override
    public boolean delete(String id) {
        return deleteFromDatastore(id);
    }

    private Action createActionInDatastore(Action action) {
        Datastore ds = dataStoreTenants.get(tenantInfo);

        Key key = createkey(ds, action.getId());
        FullEntity<Key> entity = convertActionToEntity(action, key);
        Transaction txn = ds.newTransaction();
        try {
            Entity existingEntity = txn.get(entity.getKey());
            if (existingEntity != null)
                throwConflict();
            txn.put(entity);
            txn.commit();
        } catch (DatastoreException ex) {
            if (ex.getCode() == 10) {
                if (ex.getMessage().startsWith("too much contention")) {
                    return throwConflict();
                } else {
                    throw new AppException(409, CONFLICT_REASON, "ABORTED due to conflict");
                }
            } else
                throw ex;
        } finally {
            if (txn.isActive()) {
                txn.rollback();
            }
        }
        return action;
    }

    private Action throwConflict() {
        throw new AppException(409, CONFLICT_REASON, "An action already exists with the same message and endpoint combination");
    }

    private boolean deleteFromDatastore(String id) {
        Datastore ds = dataStoreTenants.get(tenantInfo);
        boolean output = false;

        Key key = createkey(ds, id);
        Transaction txn = ds.newTransaction();
        try {
            Entity existingEntity = txn.get(key);
            if (existingEntity != null) {
                txn.delete(key);
                txn.commit();
                output = true;
            }
        } catch (DatastoreException ex) {
            if (ex.getCode() == 10) {
                if (ex.getMessage().startsWith("too much contention")) {
                    throw new AppException(409, CONFLICT_REASON, "Another request is trying to delete the same DDMS.");
                } else {
                    throw new AppException(409, CONFLICT_REASON, "ABORTED due to conflict");
                }
            } else
                throw ex;
        } finally {
            if (txn.isActive()) {
                txn.rollback();
            }
        }
        return output;

    }

    @Override
    public List<Action> getAllActions() {
        Datastore ds = dataStoreTenants.get(tenantInfo);

        Query<Entity> q = Query
                .newGqlQueryBuilder(Query.ResultType.ENTITY, "Select * from ACTION")
                .setNamespace(NAMESPACE).build();

        QueryResults<Entity> r = ds.run(q);

        List<Action> actions = new ArrayList<>();
        while (r.hasNext()) {
            Entity entity = r.next();
            Action action = convertEntityToAction(entity);
            if (getActionChecksum(action).compareTo(entity.getString(CHECKSUM)) != 0) {
                log.error(String.format("Error retrieving action '%s'.", action.getName()));
                throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "Internal server error", "Error retrieving actions");
            }
            actions.add(action);
        }
        return actions;
    }

    Key createkey(Datastore ds, String id) {
        return ds.newKeyFactory()
                .setNamespace(NAMESPACE)
                .setKind(KIND)
                .newKey(id);
    }

    Action convertEntityToAction(Entity entity) {

        List<String> types = new ArrayList<>();
        List<Value<String>> tCol = entity.getList(TYPE);
        tCol.forEach(t -> types.add(t.get()));

        List<String> sources = new ArrayList<>();
        List<Value<String>> sCol = entity.getList(SOURCE);
        sCol.forEach(t -> sources.add(t.get()));

        List<Value<String>> vCol = entity.getList(VERSION);
        List<String> versions = new ArrayList<>();
        vCol.forEach(t -> versions.add(t.get()));

        Action output = new Action();
        output.setId(entity.getKey().getName());
        output.setName(entity.getString(NAME));
        output.setDescription(entity.getString(DESCRIPTION));
        output.setUrl(entity.getString(URL));
        output.setImg(entity.getString(IMG));
        output.setContactEmail(entity.getString(CONTACT_EMAIL));
        output.setCreatedOnEpoch(entity.getTimestamp(CREATED_ON).toSqlTimestamp());
        output.getFilter().setEntityType(types);
        output.getFilter().setSource(sources);
        output.getFilter().setVersion(versions);
        return output;
    }

    private FullEntity convertActionToEntity(Action action, Key key) {
        List<StringValue> types = new ArrayList<>();
        action.getFilter().getEntityType().forEach(t -> types.add(new StringValue(t)));
        List<StringValue> source = new ArrayList<>();
        action.getFilter().getSource().forEach(t -> source.add(new StringValue(t)));
        List<StringValue> version = new ArrayList<>();
        action.getFilter().getVersion().forEach(t -> version.add(new StringValue(t)));

        FullEntity.Builder entity = FullEntity.newBuilder(key)
                .set(NAME, action.getName())
                .set(CREATED_ON, Timestamp.of(action.getCreatedOnEpoch()))
                .set(CONTACT_EMAIL, action.getContactEmail())
                .set(IMG, action.getImg())
                .set(URL, action.getUrl())
                .set(TYPE, types)
                .set(SOURCE, source)
                .set(VERSION, version)
                .set(CHECKSUM, getActionChecksum(action))
                .set(DESCRIPTION, StringValue.newBuilder(action.getDescription()).setExcludeFromIndexes(true).build());

        return entity.build();
    }

    String getActionChecksum(Action action) {
        String serializedAction = new Gson().toJson(action);
        return Crc32c.hashToBase64EncodedString(serializedAction);
    }

}
