/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.provider.gcp.subscriber;

import com.google.cloud.datastore.*;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.register.provider.gcp.ddms.datastore.DatastoreMultiTenantAccess;
import org.opengroup.osdu.register.provider.gcp.util.GcpAppServiceConfig;
import org.opengroup.osdu.register.subscriber.model.Secret;
import org.opengroup.osdu.register.subscriber.model.Subscription;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

@Repository
public class DatastoreAccess implements IDatastoreAccess {

    private static final String SERVER_ERROR_REASON = "Server Error";
    private static final String CONFLICT_REASON = "Conflict";
    private static final String TOO_MUCH_CONTENTION_MSG = "too much contention";
    private static final String ABORTED_DUE_TO_CONFLICT_MSG = "ABORTED due to conflict";

    @Autowired
    private PushSubscription pushSubscription;
    @Autowired
    private DatastoreMultiTenantAccess dataStoreTenants;
    @Autowired
    private TenantInfo tenantInfo;
    @Autowired
    private ModelEntityHelper modelEntityHelper;
    @Autowired
    private GcpAppServiceConfig config;

    private static final String NAMESPACE = "DE";
    private static final String KIND = "SUBSCRIPTION";

    @Override
    public Subscription create(Subscription s) {
        s = createSubscriberInDatastore(s);
        try {
            pushSubscription.create(s, tenantInfo, this.config.getGoogleCloudProject(), config.getServiceAccountIdentity());
        } catch (Exception e) {
            deleteFromDatastore(s.getId(), false);
            throw new AppException(500, SERVER_ERROR_REASON, "Unexpected error creating subscription", e);
        }
        return s;
    }

    @Override
    public Subscription get(String id) {
        Datastore ds = dataStoreTenants.get(tenantInfo);
        Key key = createKey(ds, id);
        Entity entity = ds.get(key);
        if (entity == null)
            throw new AppException(404, "Not found", String.format("Subscriber with id %s does not exist.", id));

        try {
            return modelEntityHelper.convertEntityToSubscription(entity);
        } catch (Exception e) {
            throw new AppException(500, SERVER_ERROR_REASON, "Unexpected error getting subscription", e);
        }
    }

    // TODO: Remove this api when switch notification to use cache
    @Override
    public List<Subscription> query(String notificationId) {
        List<Subscription> output = new ArrayList<>();
        Datastore ds = dataStoreTenants.get(tenantInfo);
        Query<Entity> q = Query
                .newGqlQueryBuilder(Query.ResultType.ENTITY, "SELECT * FROM SUBSCRIPTION WHERE notificationId=@t")
                .setBinding("t", notificationId)
                .setNamespace(NAMESPACE)
                .build();
        QueryResults<Entity> results = ds.run(q);

        try {
            while (results.hasNext()) {
                output.add(modelEntityHelper.convertEntityToSubscription(results.next()));
            }
            return output;
        } catch (Exception e) {
            throw new AppException(500, SERVER_ERROR_REASON, "Unexpected error querying subscription", e);
        }
    }

    @Override
    public boolean delete(String id) {
        return deleteFromDatastore(id, true);
    }

    @Override
    public boolean patch(String id, Secret secret) {
        return updateInDataStore(id, secret);
    }

    @Override
    public List<Subscription> getAll() throws Exception {

        Datastore ds = dataStoreTenants.get(tenantInfo);
        Query<Entity> q = Query
                .newGqlQueryBuilder(Query.ResultType.ENTITY, "Select * from SUBSCRIPTION")
                .setNamespace(NAMESPACE).build();

        QueryResults<Entity> r = ds.run(q);

        List<Subscription> subscriptions = new ArrayList<>();
        while (r.hasNext()) {
            Entity entity = r.next();
            Subscription action = modelEntityHelper.convertEntityToSubscription(entity);
            subscriptions.add(action);
        }
        return subscriptions;
    }

    private Subscription createSubscriberInDatastore(Subscription s) {
        Datastore ds = dataStoreTenants.get(tenantInfo);

        Key key = createKey(ds, s.getId());
        Transaction txn = ds.newTransaction();
        try {
            FullEntity<Key> entity = modelEntityHelper.convertSubscriptionToEntity(s, key);
            Entity existingEntity = txn.get(entity.getKey());
            if (existingEntity != null)
                throw new AppException(409, CONFLICT_REASON, "A subscriber already exists with the same topic and endpoint combination");
            txn.put(entity);
            txn.commit();
        } catch (DatastoreException ex) {
            if (ex.getCode() == 10) {
                if (ex.getMessage().startsWith(TOO_MUCH_CONTENTION_MSG)) {
                    throw new AppException(409, CONFLICT_REASON, "A subscriber already exists with the same topic and endpoint combination");
                } else {
                    throw new AppException(409, CONFLICT_REASON, ABORTED_DUE_TO_CONFLICT_MSG);
                }
            } else
                throw ex;
        } catch (IOException ex) {
            throw new AppException(500, SERVER_ERROR_REASON, "Unexpected error creating subscription.", ex);
        } finally {
            if (txn.isActive()) {
                txn.rollback();
            }
        }
        return s;
    }

    private boolean deleteFromDatastore(String id, boolean deleteSubscription) {
        Datastore ds = dataStoreTenants.get(tenantInfo);
        boolean output = false;

        Key key = createKey(ds, id);
        Transaction txn = ds.newTransaction();
        try {
            Entity existingEntity = txn.get(key);
            if (existingEntity != null) {
                if (deleteSubscription) {
                    Subscription s = modelEntityHelper.convertEntityToSubscription(existingEntity);
                    pushSubscription.delete(this.config.getGoogleCloudProject(), s.getNotificationId());
                }
                txn.delete(key);
                txn.commit();
                output = true;
            }
        } catch (DatastoreException ex) {
            if (ex.getCode() == 10) {
                if (ex.getMessage().startsWith(TOO_MUCH_CONTENTION_MSG)) {
                    throw new AppException(409, CONFLICT_REASON, "Another request is trying to delete the same DDMS.");
                } else {
                    throw new AppException(409, CONFLICT_REASON, ABORTED_DUE_TO_CONFLICT_MSG);
                }
            } else
                throw ex;
        } catch (IOException ex) {
            throw new AppException(500, SERVER_ERROR_REASON, "Unexpected error deleting subscription.", ex);
        } finally {
            if (txn.isActive()) {
                txn.rollback();
            }
        }
        return output;
    }

    private boolean updateInDataStore(String id, Secret secret) {
        Datastore ds = dataStoreTenants.get(tenantInfo);
        boolean output = false;

        Key key = createKey(ds, id);
        Transaction txn = ds.newTransaction();
        try {
            Entity existingEntity = txn.get(key);
            if (existingEntity != null) {
                Entity updatedEntity = modelEntityHelper.getUpdatedEntity(existingEntity, secret);
                txn.put(existingEntity, updatedEntity);
                txn.commit();
                output = true;
            }
        } catch (DatastoreException ex) {
            if (ex.getCode() == 10) {
                if (ex.getMessage().startsWith(TOO_MUCH_CONTENTION_MSG)) {
                    throw new AppException(409, CONFLICT_REASON, "Another request is trying to update the same subscription.");
                } else {
                    throw new AppException(409, CONFLICT_REASON, ABORTED_DUE_TO_CONFLICT_MSG);
                }
            } else
                throw ex;
        } catch (IOException ex) {
            throw new AppException(500, SERVER_ERROR_REASON, "Unexpected error updating subscription.", ex);
        } finally {
            if (txn.isActive()) {
                txn.rollback();
            }
        }
        return output;
    }

    private Key createKey(Datastore ds, String id) {
        return ds.newKeyFactory()
                .setNamespace(NAMESPACE)
                .setKind(KIND)
                .newKey(id);
    }
}
