/*
  Copyright 2021 Google LLC
  Copyright 2021 EPAM Systems, Inc

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */

package org.opengroup.osdu.register.provider.reference.subscriber;

import java.util.List;
import java.util.stream.Collectors;
import lombok.RequiredArgsConstructor;
import org.apache.http.HttpStatus;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.register.provider.interfaces.subscriber.ISubscriptionRepository;
import org.opengroup.osdu.register.provider.reference.mapper.SubscriptionMapper;
import org.opengroup.osdu.register.provider.reference.model.document.SubscriptionDocument;
import org.opengroup.osdu.register.provider.reference.repository.mongo.SubscriptionDocumentRepository;
import org.opengroup.osdu.register.subscriber.model.Secret;
import org.opengroup.osdu.register.subscriber.model.Subscription;
import org.springframework.stereotype.Repository;

@Repository
@RequiredArgsConstructor
public class ReferenceSubscriptionRepository implements ISubscriptionRepository {

  private static final String CONFLICT_REASON = "Conflict";
  private static final String CONFLICT_MESSAGE = "A subscriber already exists with the same topic and endpoint combination.";
  private static final String NOT_FOUND_REASON = "Not found";
  private static final String NOT_FOUND_MESSAGE = "Subscriber with id %s does not exist.";

  private final SubscriptionDocumentRepository subscriptionDocumentRepository;
  private final SubscriptionMapper subscriptionMapper;

  @Override
  public Subscription create(Subscription subscription) throws Exception {
    if (this.subscriptionDocumentRepository.existsById(subscription.getId())) {
      throw new AppException(HttpStatus.SC_CONFLICT, CONFLICT_REASON, CONFLICT_MESSAGE);
    }

    SubscriptionDocument subscriptionDocument = this.subscriptionMapper.toDocument(subscription);
    SubscriptionDocument saved = this.subscriptionDocumentRepository.save(subscriptionDocument);

    //TODO implement subscription creation after https://jiraeu.epam.com/browse/GONRG-3466 is resolved.
    //pushSubscription.create(subscription, ...);

    return this.subscriptionMapper.toSubscription(saved);
  }

  @Override
  public Subscription get(String id) {
    SubscriptionDocument document = this.subscriptionDocumentRepository.findById(id)
        .orElseThrow(() ->
            new AppException(HttpStatus.SC_NOT_FOUND, NOT_FOUND_REASON,
                String.format(NOT_FOUND_MESSAGE, id)));

    return this.subscriptionMapper.toSubscription(document);
  }

  @Override
  public List<Subscription> query(String notificationId) {
    List<SubscriptionDocument> documentList = this.subscriptionDocumentRepository.findByNotificationId(
        notificationId);
    return documentList.stream().map(this.subscriptionMapper::toSubscription).collect(
        Collectors.toList());
  }

  @Override
  public boolean delete(String id) {
    if (this.subscriptionDocumentRepository.existsById(id)) {
      this.subscriptionDocumentRepository.deleteById(id);
      //TODO implement subscription creation after https://jiraeu.epam.com/browse/GONRG-3466 is resolved.
      //pushSubscription.delete(subscription.getNotificationId(), ...);
      return true;
    }
    return false;
  }

  @Override
  public boolean patch(Subscription subscription, Secret secret) {
    if (this.subscriptionDocumentRepository.existsById(subscription.getId())) {
      SubscriptionDocument updatedDocument = this.subscriptionMapper.toUpdatedDocument(
          subscription, secret);
      this.subscriptionDocumentRepository.save(updatedDocument);
      return true;
    }
    return false;
  }

  @Override
  public List<Subscription> getAll() throws Exception {
    List<SubscriptionDocument> documentList = this.subscriptionDocumentRepository.findAll();
    return documentList.stream().map(this.subscriptionMapper::toSubscription)
        .collect(Collectors.toList());
  }
}
